==> sl-cia-2.0.88.tgz: package/lib/js_diff/file-signature.js <==
﻿var esprima = require('esprima'),
    escodegen = require('escodegen'),
    traverse = require('ast-traverse'),
    astUtils = require('esprima-ast-utils'),
    md5 = require('md5'),
    path = require('path'),
    guessMethodName = require('./guess-method-name.js').guessMethodName;

var branchNodeTypes = {"IfStatement": true, "SwitchStatement": true, "ConditionalExpression": true, "LogicalExpression": true};

function formatLocToArray(loc) {
    return [loc.line, loc.column];
}

function findLogicalExpressionLeaves(node, leaves) {
    if (node.type === "LogicalExpression") {
            findLogicalExpressionLeaves(node.left, leaves);
            findLogicalExpressionLeaves(node.right, leaves);
    } else {
        leaves.push(node);
    }
}

function calculateFunctionNodeSigHash(node) {
    if (!node.params || node.params.length == 0) return "";
    var paramsStr = node.params.map(escodegen.generate).join(',');
    var hash = md5(paramsStr);
    return hash;
}

function calculateFunctionNodeHash(node, isBranchCoverage) {
    var clonedNode = astUtils.clone(node);
    var branchHashes = [];
    //calculate body hash
    traverse(clonedNode, {
        pre: function (node, parent, prop, idx) {
            if (node == clonedNode) return;
            if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression") {
                node.body = { type: "BlockStatement", body: [] };
                node.params = [];
                node.id = { name: '_' };
            }
            if (isBranchCoverage) {
                if (branchNodeTypes[node.type] && node.type !== "LogicalExpression") {
                    calculateMethodBranches(node, branchHashes);
                }
                if (node.type == "LogicalExpression" && (parent == undefined || parent.type !== "LogicalExpression")) {
                    calculateMethodBranches(node, branchHashes);
                }
            }
        }
    });

    var generated = escodegen.generate(clonedNode);
    var hash = md5(generated);
    if (isBranchCoverage) {
        return {hash: hash, branchHashes: branchHashes}
    }
    return  {hash: hash}
}

/**
 * Taken from eslint - https://github.com/eslint/eslint/blob/183def6115cad6f17c82ef1c1a245eb22d0bee83/lib/eslint.js#L800
 */
function trimShebang(text) {
  return text.replace(/^#!([^\r\n]+)/, function(match, captured) { return "//" + captured; });
}

function calculateMethodBranches(node, branches) {
    switch (node.type) {
        case "IfStatement":
        case "ConditionalExpression":
            if (node.consequent !== null) {
                branches.push(calculateSingleBranchHash(node.consequent, node.type));
            }
            if (node.alternate == null) {
                // Istanbul counts empty alternates and takes the consequent's location
                var clonedNode = astUtils.clone(node);
                clonedNode.alternate = {type: "BlockStatement", body: []};
                clonedNode.alternate.loc = node.consequent !== null ? node.consequent.loc : node.loc;
                branches.push(calculateSingleBranchHash(clonedNode.alternate, node.type));
            }
            else {
                branches.push(calculateSingleBranchHash(node.alternate, node.type));
            }
            break;
        case "SwitchStatement":
            node.cases.forEach(function (caseNode) {
                if (caseNode !== null) {
                    branches.push(calculateSingleBranchHash(caseNode, node.type));
                }
            });
            break;
        case "LogicalExpression":
            var leaves = [];
            findLogicalExpressionLeaves(node, leaves);
            leaves.forEach(function (leafNode) {
                if (leafNode !== null) {
                    branches.push(calculateSingleBranchHash(leafNode, node.type));
                }
            });
            break;
    }
}

function calculateSingleBranchHash(node, nodeType) {
    var clonedNode = astUtils.clone(node);
    var position = formatLocToArray(node.loc.start);
    traverse(clonedNode, {
        pre: function (node, parent, prop, idx) {
            if (node == clonedNode) return;
            if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression") {
                node.body = { type: "BlockStatement", body: [] };
                node.params = [];
                node.id = { name: '_' };
            }
            switch (node.type) {
                case "IfStatement":
                case "ConditionalExpression":
                    node.consequent = {type: "BlockStatement", body: []};
                    node.alternate = {type: "BlockStatement", body: []};
                    break;
                case "SwitchStatement":
                    node.cases = [];
                    break;
                case "LogicalExpression":
                    node.operator = "";
                    node.left = {type: "Literal", value: null};
                    node.right = {type: "Literal", value: null};
                    break;
            }
        }
    });
    var generated = escodegen.generate(clonedNode);
    var hash = md5(generated);
    return {
        type: nodeType,
        position: position,
        hash: hash
    };
}

function getMethodName(node, parent) {
    var mn = node.id && node.id.name;
    if (!mn) {
        mn = guessMethodName(node, parent) || "(Anonymous)";
    }
    return mn;
}

function fixRelPath(p1, p2) {
    try {
        if (!p1 || p1.length == 0) return p1;
        if (!p2 || p2.length == 0) return p1;

        //Make sure both paths are unix-style
        p1 = p1.split('\\').join('/');
        p2 = p2.split('\\').join('/');

        if (p2[0] == '/') return p2; //Absolute path

        var parts1 = p1.split('/');
        parts1.pop(); //remove the filename

        var parts2 = p2.split('/');

        var p3 = parts1.concat(parts2);
        var finalPath = path.normalize(p3.join('/')).replace('\\', '/');

        return finalPath;
    } catch (e) {
        console.error(e);
        return p2;
    }
}

function generateFileSignature(filename, source, sourceMaps, isBranchCoverage) {
    var result = {
        module: { hash: "" },
        methods: []
    };

    filename = filename.replace(/\\/g, '/'); //we want all transients to use unix / instead of windows \

    var ast;

    try {
        source = trimShebang(source);

        //1. Parse the file
        ast = esprima.parse(source, {
            loc: true,
            range: true,
            comment: false //No need for those
        });

    } catch (parseException) {
        //TODO: Test
        return { error: "Parse error (1)" };
    }

    try {

        // print AST node types, pre-order (node first, then its children) 
        traverse(ast, {
            pre: function (node, parent, prop, idx) {
                if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression") {
                    node.isFunctionNode = true;
                    var methodHashObject = calculateFunctionNodeHash(node, isBranchCoverage);
                    var hash = methodHashObject.hash;
                    var sigHash = calculateFunctionNodeSigHash(node);
                    var start = formatLocToArray(node.loc.start);
                    var end = formatLocToArray(node.loc.end);
                    var name = getMethodName(node, parent);
                    var fn = filename;
                    var originalFilename;
                    var originalName;
                    var mappedPosition;
                    var originalStart;
                    var originalEnd;
                    var originalStartPosition, originalEndPosition;

                    if (sourceMaps) {
                        try {
                            originalStartPosition = sourceMaps.originalPositionFor({ line: start[0], column: start[1] });
                            if (!originalStartPosition || !originalStartPosition.source || originalStartPosition.line === null || originalStartPosition.column === null) {
                                return;
                            }
                            originalFilename = fn;
                            fn = originalStartPosition.source;

                            if (fn.indexOf("webpack:/webpack/") === 0) return; //Embedded resource that has no corresponding source
                            if (fn.indexOf("webpack:///external") === 0) return; //Ignore external module
                            if (fn.indexOf('webpack:///') === 0) {
                                fn = fn.substring(11); //'webpack:///'.length
                            }

                            fn = fixRelPath(originalFilename, fn);

                            if (originalStartPosition.name) {
                                originalName = node.id && node.id.name;
                                name = originalStartPosition.name;
                                if (!node.id)
                                    node.id = { name: originalStartPosition.name };
                                else if (!node.id.name)
                                    node.id.name = originalStartPosition.name;
                            }

                            originalStart = start;
                            start = [originalStartPosition.line, originalStartPosition.column];

                            originalEndPosition = sourceMaps.originalPositionFor({ line: end[0], column: end[1] - 1 });
                            originalEnd = end;
                            end = [originalEndPosition.line, originalEndPosition.column];

                        } catch (e) {
                            console.log('Error calculating source maps:' + e);
                        }
                    }

                    var hasExplicitName = (node.id && node.id.name);
                    var uniqueName = [name, fn, start.join(',')].join('@');

                    var method = {
                        name: name, //Method that can be presented to the user
                        isAnonymous: !hasExplicitName,
                        start: start,
                        end: end,
                        hash: hash,
                        sigHash: sigHash,
                        uniqueName: uniqueName,
                        filename: fn,
                        originalFilename: originalFilename,
                        originalStart: originalStart,
                        originalEnd: originalEnd,
                        originalName: originalName
                    };
                    if (isBranchCoverage) {
                        method.branches = methodHashObject.branchHashes;
                    }
                    result.methods.push(method);
                }
                if (node.hasOwnProperty("isFunctionNode") ||
                    (parent !== null && parent.hasOwnProperty("isFunctionNode"))) {
                    // We want to tag nodes that have a parent that is a method
                    node.isFunctionNode = true;
                }
                if (branchNodeTypes[node.type] &&
                    (!node.hasOwnProperty("isFunctionNode") || node.isFunctionNode !== true)) {
                    // Nodes that are of branch types and do not belong to a method parent are global branches
                    //TODO Global branches
                }
            }
        });
    }
    catch (traverseException1) {
        return { error: 'Parse error(2)' }
    }

    try {
        var bodyHash = calculateFunctionNodeHash(ast).hash;
        result.module.hash = bodyHash;
    } catch (traverseException2) {
        return { error: "Parse error (3)" };
    }

    return result;
}

module.exports = generateFileSignature;

==> sl-cia-2.0.97.tgz: package/lib/js_diff/file-signature.js <==
﻿var esprima = require('esprima'),
    escodegen = require('escodegen'),
    traverse = require('ast-traverse'),
    astUtils = require('esprima-ast-utils'),
    md5 = require('md5'),
    path = require('path'),
    guessMethodName = require('./guess-method-name.js').guessMethodName,
    consoleFileLogger = require("../sl-logger.js").consoleFileLogger;

var branchNodeTypes = {"IfStatement": true, "SwitchStatement": true, "ConditionalExpression": true, "LogicalExpression": true};

function formatLocToArray(loc) {
    return [loc.line, loc.column];
}

function findLogicalExpressionLeaves(node, leaves) {
    if (node.type === "LogicalExpression") {
            findLogicalExpressionLeaves(node.left, leaves);
            findLogicalExpressionLeaves(node.right, leaves);
    } else {
        leaves.push(node);
    }
}

function calculateFunctionNodeSigHash(node) {
    if (!node.params || node.params.length == 0) return "";
    var paramsStr = node.params.map(escodegen.generate).join(',');
    var hash = md5(paramsStr);
    return hash;
}

function calculateFunctionNodeHash(node, isBranchCoverage) {
    var clonedNode = astUtils.clone(node);
    var branchHashes = [];
    //calculate body hash
    traverse(clonedNode, {
        pre: function (node, parent, prop, idx) {
            if (node == clonedNode) return;
            if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression") {
                node.body = { type: "BlockStatement", body: [] };
                node.params = [];
                node.id = { name: '_' };
            }
            if (isBranchCoverage) {
                if (branchNodeTypes[node.type] && node.type !== "LogicalExpression") {
                    calculateMethodBranches(node, branchHashes);
                }
                if (node.type == "LogicalExpression" && (parent == undefined || parent.type !== "LogicalExpression")) {
                    calculateMethodBranches(node, branchHashes);
                }
            }
        }
    });

    var generated = escodegen.generate(clonedNode);
    var hash = md5(generated);
    if (isBranchCoverage) {
        return {hash: hash, branchHashes: branchHashes}
    }
    return  {hash: hash}
}

/**
 * Taken from eslint - https://github.com/eslint/eslint/blob/183def6115cad6f17c82ef1c1a245eb22d0bee83/lib/eslint.js#L800
 */
function trimShebang(text) {
  return text.replace(/^#!([^\r\n]+)/, function(match, captured) { return "//" + captured; });
}

function calculateMethodBranches(node, branches) {
    switch (node.type) {
        case "IfStatement":
        case "ConditionalExpression":
            if (node.consequent !== null) {
                branches.push(calculateSingleBranchHash(node.consequent, node.type));
            }
            if (node.alternate == null) {
                // Istanbul counts empty alternates and takes the consequent's location
                var clonedNode = astUtils.clone(node);
                clonedNode.alternate = {type: "BlockStatement", body: []};
                clonedNode.alternate.loc = node.consequent !== null ? node.consequent.loc : node.loc;
                branches.push(calculateSingleBranchHash(clonedNode.alternate, node.type));
            }
            else {
                branches.push(calculateSingleBranchHash(node.alternate, node.type));
            }
            break;
        case "SwitchStatement":
            node.cases.forEach(function (caseNode) {
                if (caseNode !== null) {
                    branches.push(calculateSingleBranchHash(caseNode, node.type));
                }
            });
            break;
        case "LogicalExpression":
            var leaves = [];
            findLogicalExpressionLeaves(node, leaves);
            leaves.forEach(function (leafNode) {
                if (leafNode !== null) {
                    branches.push(calculateSingleBranchHash(leafNode, node.type));
                }
            });
            break;
    }
}

function calculateSingleBranchHash(node, nodeType) {
    var clonedNode = astUtils.clone(node);
    var position = formatLocToArray(node.loc.start);
    traverse(clonedNode, {
        pre: function (node, parent, prop, idx) {
            if (node == clonedNode) return;
            if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression") {
                node.body = { type: "BlockStatement", body: [] };
                node.params = [];
                node.id = { name: '_' };
            }
            switch (node.type) {
                case "IfStatement":
                case "ConditionalExpression":
                    node.consequent = {type: "BlockStatement", body: []};
                    node.alternate = {type: "BlockStatement", body: []};
                    break;
                case "SwitchStatement":
                    node.cases = [];
                    break;
                case "LogicalExpression":
                    node.operator = "";
                    node.left = {type: "Literal", value: null};
                    node.right = {type: "Literal", value: null};
                    break;
            }
        }
    });
    var generated = escodegen.generate(clonedNode);
    var hash = md5(generated);
    return {
        type: nodeType,
        position: position,
        hash: hash
    };
}

function getMethodName(node, parent) {
    var mn = node.id && node.id.name;
    if (!mn) {
        mn = guessMethodName(node, parent) || "(Anonymous)";
    }
    return mn;
}

function fixRelPath(p1, p2, logger) {
    var log = logger? logger.child({methodName: "fixRelPath"}) : consoleFileLogger("fixRelPath");
    try {
        if (!p1 || p1.length == 0) return p1;
        if (!p2 || p2.length == 0) return p1;

        //Make sure both paths are unix-style
        p1 = p1.split('\\').join('/');
        p2 = p2.split('\\').join('/');

        if (p2[0] == '/') return p2; //Absolute path

        var parts1 = p1.split('/');
        parts1.pop(); //remove the filename

        var parts2 = p2.split('/');

        var p3 = parts1.concat(parts2);
        var finalPath = path.normalize(p3.join('/')).replace('\\', '/');

        return finalPath;
    } catch (err) {
        log.error("Failed to fix real path for source maps. Error: " + err + ". Original File Name: " + p1 + ". Function: " + p2);
        return p2;
    }
}

function generateFileSignature(filename, source, sourceMaps, isBranchCoverage, logger) {
    var result = {
        module: { hash: "" },
        methods: []
    };
    var log = logger? logger.child({methodName: "generateFileSignature"}) : consoleFileLogger("generateFileSignature");

    filename = filename.replace(/\\/g, '/'); //we want all transients to use unix / instead of windows \

    var ast;

    try {
        source = trimShebang(source);

        //1. Parse the file
        ast = esprima.parse(source, {
            loc: true,
            range: true,
            comment: false //No need for those
        });

    } catch (parseException) {
        //TODO: Test
        log.error("Failed to parse filename to ast. Filename: " + filename + ". Error: " + parseException);
        return { error: "Parse error (1)" };
    }

    try {

        // print AST node types, pre-order (node first, then its children) 
        traverse(ast, {
            pre: function (node, parent, prop, idx) {
                if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression") {
                    node.isFunctionNode = true;
                    var methodHashObject = calculateFunctionNodeHash(node, isBranchCoverage);
                    var hash = methodHashObject.hash;
                    var sigHash = calculateFunctionNodeSigHash(node);
                    var start = formatLocToArray(node.loc.start);
                    var end = formatLocToArray(node.loc.end);
                    var name = getMethodName(node, parent);
                    var fn = filename;
                    var originalFilename;
                    var originalName;
                    var mappedPosition;
                    var originalStart;
                    var originalEnd;
                    var originalStartPosition, originalEndPosition;

                    if (sourceMaps) {
                        try {
                            originalStartPosition = sourceMaps.originalPositionFor({ line: start[0], column: start[1] });
                            if (!originalStartPosition || !originalStartPosition.source || originalStartPosition.line === null || originalStartPosition.column === null) {
                                return;
                            }
                            originalFilename = fn;
                            fn = originalStartPosition.source;

                            if (fn.indexOf("webpack:/webpack/") === 0) return; //Embedded resource that has no corresponding source
                            if (fn.indexOf("webpack:///external") === 0) return; //Ignore external module
                            if (fn.indexOf('webpack:///') === 0) {
                                fn = fn.substring(11); //'webpack:///'.length
                            }

                            fn = fixRelPath(originalFilename, fn, log);

                            if (originalStartPosition.name) {
                                originalName = node.id && node.id.name;
                                name = originalStartPosition.name;
                                if (!node.id)
                                    node.id = { name: originalStartPosition.name };
                                else if (!node.id.name)
                                    node.id.name = originalStartPosition.name;
                            }

                            originalStart = start;
                            start = [originalStartPosition.line, originalStartPosition.column];

                            originalEndPosition = sourceMaps.originalPositionFor({ line: end[0], column: end[1] - 1 });
                            originalEnd = end;
                            end = [originalEndPosition.line, originalEndPosition.column];

                        } catch (e) {
                            log.error('Error calculating source maps:' + e);
                        }
                    }

                    var hasExplicitName = (node.id && node.id.name);
                    var uniqueName = [name, fn, start.join(',')].join('@');

                    var method = {
                        name: name, //Method that can be presented to the user
                        isAnonymous: !hasExplicitName,
                        start: start,
                        end: end,
                        hash: hash,
                        sigHash: sigHash,
                        uniqueName: uniqueName,
                        filename: fn,
                        originalFilename: originalFilename,
                        originalStart: originalStart,
                        originalEnd: originalEnd,
                        originalName: originalName
                    };
                    if (isBranchCoverage) {
                        if (sourceMaps) {
                            methodHashObject.branchHashes.forEach(function (branchHash) {
                                var originalBranchPosition = sourceMaps.originalPositionFor(
                                    { line: branchHash.position[0], column: branchHash.position[1] }
                                );
                                if (!originalBranchPosition || originalBranchPosition.line === null || originalBranchPosition.column === null)
                                {
                                    return;
                                }
                                branchHash.position = formatLocToArray(originalBranchPosition);
                            })
                        }
                        method.branches = methodHashObject.branchHashes;
                    }
                    result.methods.push(method);
                }
                if (node.hasOwnProperty("isFunctionNode") ||
                    (parent !== null && parent.hasOwnProperty("isFunctionNode"))) {
                    // We want to tag nodes that have a parent that is a method
                    node.isFunctionNode = true;
                }
                if (branchNodeTypes[node.type] &&
                    (!node.hasOwnProperty("isFunctionNode") || node.isFunctionNode !== true)) {
                    // Nodes that are of branch types and do not belong to a method parent are global branches
                    //TODO Global branches
                }
            }
        });
    }
    catch (traverseException1) {
        log.error("Failed to traverse ast. Filename: " + filename + ". Error: " + traverseException1);
        return { error: 'Parse error(2)' }
    }

    try {
        var bodyHash = calculateFunctionNodeHash(ast).hash;
        result.module.hash = bodyHash;
    } catch (traverseException2) {
        log.error("Failed to calculate module hash. Filename: " + filename + ". Error: traverseException2");
        return { error: "Parse error (3)" };
    }

    return result;
}

module.exports = generateFileSignature;
